#include <pthread.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <curl/curl.h>
#include <stdio.h>
#include "debug.h"
#include "md5.h"
#include "queue.h"
#include "hatena.h"
#include "config.h"

#define HATENA_HS_URL "http://music.hatelabo.jp/trackauth"
#define HATENA_CLI_ID "aud"
#define HATENA_HS_WAIT 1800
#define HATENA_SB_WAIT 10
#define HATENA_VERSION "1.1"
/* This is the implementation, not the player version. */
#define HATENA_IMPLEMENTATION "0.1"
#define HATENA_SB_MAXLEN 1024

#define SC_CURL_TIMEOUT 15
#define BADAUTH_LIMIT 1

#define MULTIPLE_SUBMISSION_MAX 10

#define DEFAULT_SUBMIT_INTERVAL 60

/* Scrobblerbackend for xmms plugin, first draft */

/* for debug */
#define DEBUG_HS_CURL 0

#define DEBUG_HS_RESPONSE_EMPTY 0
#define DEBUG_HS_RESPONSE_FAILED 0  // FAILED
#define DEBUG_HS_RESPONSE_UPDATE 0
#define DEBUG_HS_RESPONSE_BADUSER 0
#define DEBUG_HS_RESPONSE_UNKNOWN 0

#define DEBUG_HS_BADRESPONSE 0

#define DEBUG_HS_RESPONSE_UPTODATE ((! DEBUG_HS_RESPONSE_EMPTY	  \
									 && ! DEBUG_HS_RESPONSE_FAILED	\
									 && ! DEBUG_HS_RESPONSE_UPDATE	\
									 && ! DEBUG_HS_RESPONSE_BADUSER \
									 && ! DEBUG_HS_RESPONSE_UNKNOWN) \
									|| DEBUG_HS_BADRESPONSE)

#define DEBUG_HS_NOUSERNAME 0
#define DEBUG_HS_NOPASSWORD 0


#define DEBUG_SB_CURL 0

#define DEBUG_SB_RESPONSE_EMPTY 0
#define DEBUG_SB_RESPONSE_BADAUTH 0
#define DEBUG_SB_RESPONSE_FAILED 0
#define DEBUG_SB_RESPONSE_UNKNOWN 0
#define DEBUG_SB_RESPONSE_OK (! DEBUG_SB_RESPONSE_EMPTY \
							  && ! DEBUG_SB_RESPONSE_BADAUTH \
							  && ! DEBUG_SB_RESPONSE_FAILED \
							  && ! DEBUG_SB_RESPONSE_UNKNOWN)




static char *sc_srv_res = NULL;      // http receive buffer
static int sc_srv_res_size = 0;      // size of current receive buffer

static int sc_hs_status = 0;         // handshake flag
static int sc_hs_wait_until = 0;     // handshake retry timer
static int sc_hs_errors = 0;         // handshake error counter

static int sc_sb_errors = 0;         // submission error counter
/* static int sc_submit_interval = DEFAULT_SUBMIT_INTERVAL; */
static int sc_sb_wait_until = 0;

static int sc_badauth_count = 0;

static int sc_giveup = 0;            // gives up the transaction
static int sc_must_update = 0;       // client must be updated

static int sc_use_proxy = 0;
static int sc_use_proxy_auth = 0;
static char *sc_proxy_server = NULL;
static char *sc_proxy_userpwd = NULL;

static char *sc_username = NULL;
static char *sc_password = NULL;
static char *sc_submit_url = NULL;      // URL for submitting
static char *sc_challenge_hash = NULL;  // hash for authentication

static char sc_response_hash[33] = {0};
static char sc_curl_errbuf[CURL_ERROR_SIZE] = {0};

static pthread_mutex_t sc_status_mutex; 
static sc_status_code sc_status = NOT_RUNNING;
static CURLcode sc_curl_status = CURLE_OK;


static pthread_mutex_t sc_queue_mutex;

static void (* cb_status_update) (sc_status_code) = NULL;

static char *sc_cachefile = NULL;


static sc_entry_t *
sc_new_entry(char *artist, char *title, char len[16], 
			 char *utctime, char *album, char *mb)
{
	sc_entry_t *entry;

	entry = (sc_entry_t *)malloc(sizeof(sc_entry_t));
	
	entry->artist = artist ? curl_escape(artist, 0) : NULL;
	entry->title = title ? curl_escape(title, 0) : NULL;

	if (len && strlen(len) <= sizeof(entry->len)) {
		strncpy(entry->len, len, sizeof(len));
		entry->len[sizeof(len) - 1] = '\0';
	} else {
		memset(entry->len, 0, sizeof(entry->len));
	}

	entry->utctime = utctime ? curl_escape(utctime, 0) : NULL;

	entry->album = album ? curl_escape(album, 0) : NULL;

	entry->mb = mb ? curl_escape(mb, 0) : NULL;

	return entry;
}


static void
sc_delete_entry(void *entry)
{
	sc_entry_t *sc_entry = (sc_entry_t *)entry;

	if (sc_entry->artist) curl_free(sc_entry->artist);
	if (sc_entry->title) curl_free(sc_entry->title);
	memset(sc_entry->len, 0, sizeof(sc_entry->len));

	if (sc_entry->utctime) curl_free(sc_entry->utctime);
	if (sc_entry->album) curl_free(sc_entry->album);
	if (sc_entry->mb) curl_free(sc_entry->mb);
	
	free(sc_entry);
}


static void
sc_update_status(sc_status_code status)
{
	/*
	if (cb_status_update)
		cb_status_update(status);
	*/
	return;
}

static void
sc_set_status(sc_status_code status)
{
	//pdebug("mutex lock", DEBUG);
	pthread_mutex_lock(&sc_status_mutex);

	sc_status = status;

	pthread_mutex_unlock(&sc_status_mutex);
}

static void
sc_set_status_curl(CURLcode curl_status)
{
	//pdebug("mutex lock", DEBUG);
	pthread_mutex_lock(&sc_status_mutex);

	sc_curl_status = curl_status;

	pthread_mutex_unlock(&sc_status_mutex);
}


static size_t
sc_store_res(void *ptr, size_t size, size_t nmemb, 
			 void *stream __attribute__((unused)))
{
	int len = size * nmemb;

	sc_srv_res = realloc(sc_srv_res, sc_srv_res_size + len + 1);
	memcpy(sc_srv_res + sc_srv_res_size, ptr, len);
	sc_srv_res_size += len;
	return len;
}

static void
sc_free_res(void)
{
	if (sc_srv_res != NULL) free(sc_srv_res);
	sc_srv_res = NULL;
	sc_srv_res_size = 0;
}

static sc_status_code
sc_parse_hs_res(void)
{
	char *interval;

	if (! sc_srv_res_size || DEBUG_HS_RESPONSE_EMPTY) {
		/* non serious error */
		pdebug("No reply from server", DEBUG);
		
		return HS_EMPTY_BODY;
	}

	*(sc_srv_res + sc_srv_res_size) = '\0';

	if (! strncmp(sc_srv_res, "FAILED ", strlen("FAILED "))
		|| DEBUG_HS_RESPONSE_FAILED) {

		/* Server error? */
		/*
		interval = strstr(sc_srv_res, "INTERVAL");

		if (! interval) {
			pdebug("missing INTERVAL", DEBUG);
		} else {
			*(interval - 1) = '\0';
			sc_submit_interval 
			  = strtol(interval + strlen("INTERVAL"), NULL, 10);
		}
		*/

		pdebug(ht_fmt_vastr("error: %s", sc_srv_res), DEBUG);

		return HS_SERVER_FAILED;

	} else if (! strncmp(sc_srv_res, "UPDATE ", strlen("UPDATE "))
			   || DEBUG_HS_RESPONSE_UPDATE) {

		interval = strstr(sc_srv_res, "INTERVAL");

		/*
		if(! interval) {
			pdebug("missing INTERVAL", DEBUG);
		} else {
			*(interval - 1) = '\0';
			sc_submit_interval 
			  = strtol(interval + strlen("INTERVAL"), NULL, 10);
		}
		*/

		/* Throwing major error. Need to alert client to update. */

		pdebug(ht_fmt_vastr("update client: %s", 
							sc_srv_res + strlen("UPDATE ")), DEBUG);
		
		/*
		 * Russ isn't clear on whether we can submit with a not-updated
		 * client.  Neither is RJ.  I use what we did before.
		 */
		sc_giveup = 1;
		sc_must_update = 1;

		return HS_UPDATE_PLUGIN;

	} else if (! strncmp(sc_srv_res, "UPTODATE\n", strlen("UPTODATE\n"))
			   && DEBUG_HS_RESPONSE_UPTODATE) {

		char *p1, *p2;

		/*
		interval = strstr(sc_srv_res, "INTERVAL");

		if (! interval) {

			pdebug("missing INTERVAL", DEBUG);
		*/
			/*
			 * This is probably a bad thing, but Russ seems to
			 * think its OK to assume that an UPTODATE response
			 * may not have an INTERVAL...  We return -1 anyway.
			 */
		/*
			return HS_NOINTERVAL;
		} else {
			*(interval - 1) = '\0';
			sc_submit_interval 
			  = strtol(interval + strlen("INTERVAL"), NULL, 10); 
		}
		*/

		p1 = sc_srv_res; /* skip 1st line */

		if (! (p2 = strchr(p1, '\n'))
			|| DEBUG_HS_BADRESPONSE)
			return HS_BROKEN_RESPONSE;

		p1 = p2 + 1;

		if (*p1 == '\0' || ! (p2 = strchr(p1, '\n'))
			|| DEBUG_HS_BADRESPONSE)
			return HS_BROKEN_RESPONSE;
		
		*p2 = '\0';

		if (sc_challenge_hash) free(sc_challenge_hash);

		sc_challenge_hash = strdup(p1);

		p1 = p2 + 1;

		if (*p1 == '\0' || ! (p2 = strchr(p1, '\n'))
			|| DEBUG_HS_BADRESPONSE)
			return HS_BROKEN_RESPONSE;

		*p2 = '\0';

		if (sc_submit_url) free(sc_submit_url);

		sc_submit_url = strdup(p1);

		/*
		sc_challenge_hash = strchr(sc_srv_res, '\n') + 1;
		*(sc_challenge_hash - 1) = '\0';
		sc_challenge_hash = strdup(sc_challenge_hash);

		sc_submit_url = strchr(strchr(sc_srv_res, '\n') + 1, '\n') + 1;
		*(sc_submit_url - 1) = '\0';
		sc_submit_url = strdup(sc_submit_url);
		*/

		return HS_OK;

	} else if (! strncmp(sc_srv_res, "BADUSER", strlen("BADUSER")) 
			   || DEBUG_HS_RESPONSE_BADUSER) {

		pdebug("incorrect username/password", DEBUG);

		/*
		interval = strstr(sc_srv_res, "INTERVAL");
		if (! interval) {
			pdebug("missing INTERVAL", DEBUG);
		} else {
			*(interval - 1) = 0;
			sc_submit_interval
			  = strtol(interval + strlen("INTERVAL"), NULL, 10); 
		}
		*/

		sc_giveup = 1;
		
		return HS_BADUSERNAME;

	} else {

		pdebug(ht_fmt_vastr("unknown server response '%s'", sc_srv_res), DEBUG);

		return HS_UNKNOWN_ERROR;

	}
}

static void
hexify(char *dst, char *src, int len)
{
	//	char *bp = sc_response_hash;
	char *bp = dst;
	char hexchars[] = "0123456789abcdef";
	int i;

	//memset(sc_response_hash, 0, sizeof(sc_response_hash));
	
	for(i = 0; i < len; i ++) {
		*(bp ++) = hexchars[(src[i] >> 4) & 0x0f];
		*(bp ++) = hexchars[src[i] & 0x0f];
	}
	*bp = '\0';
}


static sc_status_code
sc_handshake(void)
{
	int status;
	char buf[4096];
	CURL *curl;
	sc_status_code ret;

	if (! sc_username || strlen(sc_username) == 0 
		|| DEBUG_HS_NOUSERNAME) {
		sc_giveup = 1;
		return HS_NOUSERNAME;
	}

	if (! sc_password || strlen(sc_password) == 0
		|| DEBUG_HS_NOPASSWORD) {
		sc_giveup = 1;
		return HS_NOPASSWORD;
	}


	snprintf(buf, sizeof(buf), "%s/?hs=true&p=%s&c=%s&v=%s&u=%s",
			 HATENA_HS_URL, HATENA_VERSION,
			 HATENA_CLI_ID, HATENA_IMPLEMENTATION, sc_username);

	pdebug(ht_fmt_vastr("Username: %s", sc_username), DEBUG);

	curl = curl_easy_init();
	curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1);
	curl_easy_setopt(curl, CURLOPT_URL, buf);
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, sc_store_res);
	memset(sc_curl_errbuf, 0, sizeof(sc_curl_errbuf));
	curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, sc_curl_errbuf);
	curl_easy_setopt(curl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_0);
	curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, SC_CURL_TIMEOUT);

	if (sc_use_proxy) {
		curl_easy_setopt(curl, CURLOPT_PROXY, sc_proxy_server);

		if (sc_use_proxy_auth)
			curl_easy_setopt(curl, CURLOPT_PROXYUSERPWD, sc_proxy_userpwd);
	}

	sc_set_status(HS_CURL_CONNECT);
	sc_update_status(HS_CURL_CONNECT);

	pdebug("Now connecting...", DEBUG);

	status = DEBUG_HS_CURL ? CURLE_COULDNT_CONNECT 
		: curl_easy_perform(curl);

	curl_easy_cleanup(curl);


	if (status) {

		pdebug(sc_curl_errbuf, DEBUG);
		sc_hs_errors ++;
		sc_free_res();

		sc_set_status_curl(status);

		return HS_CURL_ERROR;
	}

	if ((ret = sc_parse_hs_res()) != HS_OK) {
		sc_hs_errors ++;
		sc_free_res();
		return ret;
	}

	if (sc_challenge_hash != NULL) {
		md5_state_t md5state;
		unsigned char md5pword[16];
		char md5pword_str[16 * 2 + 1];

		hatena_md5_init(&md5state);
		hatena_md5_append(&md5state, (unsigned const char *)sc_password,
				   strlen(sc_password));

		hatena_md5_finish(&md5state, md5pword);

		hexify(md5pword_str, (char*)md5pword, sizeof(md5pword));
		
		hatena_md5_init(&md5state);

		hatena_md5_append(&md5state, (unsigned const char *)md5pword_str,
				   strlen(md5pword_str));
		
		hatena_md5_append(&md5state,
							 (unsigned const char *)sc_challenge_hash,
				   strlen(sc_challenge_hash));
		hatena_md5_finish(&md5state, md5pword);

		hexify(sc_response_hash, (char*)md5pword, sizeof(md5pword));

	}

	sc_hs_errors = 0;

	sc_free_res();

	return ret;
}



static sc_status_code
sc_parse_sb_res(void)
{
	/* char *ch; */

	if (! sc_srv_res_size
		|| DEBUG_SB_RESPONSE_EMPTY) {

		pdebug("No response from server", DEBUG);

		return SB_EMPTY_BODY;
	}

	*(sc_srv_res + sc_srv_res_size) = '\0';

	if (! strncmp(sc_srv_res, "OK", strlen("OK"))
		&& DEBUG_SB_RESPONSE_OK) {

		/*
		if ((ch = strstr(sc_srv_res, "INTERVAL"))) {
			sc_submit_interval = strtol(ch + strlen("INTERVAL"), NULL, 10);
			pdebug(ht_fmt_vastr("got new interval: %d", sc_submit_interval),
				   DEBUG);
		}
		*/

		pdebug(ht_fmt_vastr("submission ok: %s", sc_srv_res), DEBUG);

		return SB_OK;

	} else if (! strncmp(sc_srv_res, "BADAUTH", strlen("BADAUTH"))
			   || DEBUG_SB_RESPONSE_BADAUTH) {

		/*
		if ((ch = strstr(sc_srv_res, "INTERVAL"))) {
			sc_submit_interval = strtol(ch + strlen("INTERVAL"), NULL, 10);
			pdebug(ht_fmt_vastr("got new interval: %d", sc_submit_interval),
				   DEBUG);
		}
		*/

		pdebug("incorrect username/password", DEBUG);

		/*
		 * We obviously aren't authenticated.  The server might have
		 * lost our handshake status though, so let's try
		 * re-handshaking...  This might not be proper.
		 * (we don't give up)
		 */
		sc_hs_status = 0;

		if(sc_challenge_hash != NULL) free(sc_challenge_hash);
		sc_challenge_hash = NULL;

		if(sc_submit_url != NULL) free(sc_submit_url);
		sc_submit_url = NULL;

		pdebug("BADAUTH returns on submission. Halting "
			   "submissions until login fixed.", DEBUG);

		sc_badauth_count ++;
		if (sc_badauth_count > BADAUTH_LIMIT) {
			sc_giveup = 1;
			return SB_BADAUTHENTICATON;
		} else {
			return SB_BADAUTHENTICATON_1ST;
		}

	} else if (! strncmp(sc_srv_res, "FAILED", strlen("FAILED")) 
			   || DEBUG_SB_RESPONSE_FAILED) {

		/*
		if ((ch = strstr(sc_srv_res, "INTERVAL"))) {
			sc_submit_interval = strtol(ch + strlen("INTERVAL"), NULL, 10);
			pdebug(ht_fmt_vastr("got new interval: %d", sc_submit_interval), 
				   DEBUG);
		}
		*/

		/* This could be important. (Such as FAILED - Get new plugin) */

		pdebug(sc_srv_res, DEBUG);

		return SB_SERVER_FAILED;

	} else {

		pdebug(ht_fmt_vastr("unknown server-reply %s", sc_srv_res), DEBUG);

		return SB_UNKNOWN_ERROR;
	}
}



static void
sc_generate_entry_item(char **submission, char tag, int num, char *val)
{
    static char buf[HATENA_SB_MAXLEN]; 
	size_t submission_size;

	if (*submission == NULL)
		submission_size = 0;
	else
		submission_size = strlen(*submission);

    snprintf(buf, HATENA_SB_MAXLEN, "&%c[%d]=%s", tag, num, val ? val : "");

	*submission = (char *)realloc(*submission, 
								  submission_size + strlen(buf) + 1);
	if (submission_size == 0)
		*submission[0] = '\0';

	strncat(*submission, buf, strlen(buf));
}


static int
sc_generate_entry(char **submission)
{
	int i;
	sc_entry_t *entry;
	
	i = 0;
	*submission = NULL;

#ifdef ALLOW_MULTIPLE

	hatena_queue_peek_all_rewind();

	while (i < MULTIPLE_SUBMISSION_MAX
		   && (entry = (sc_entry_t *)hatena_queue_peek_all())) {
#else
		entry = hatena_queue_peek();
#endif
		if (! entry)
			return i;

		sc_generate_entry_item(submission, 'a', i, entry->artist);
		sc_generate_entry_item(submission, 't', i, entry->title);
		sc_generate_entry_item(submission, 'l', i, entry->len);
		sc_generate_entry_item(submission, 'i', i, entry->utctime);
		sc_generate_entry_item(submission, 'm', i, entry->mb);
		sc_generate_entry_item(submission, 'b', i, entry->album);

#ifdef ALLOW_MULTIPLE
		i++;
	}
#endif

	if (*submission)
		pdebug(ht_fmt_vastr("%s", *submission), DEBUG);

	return i;
}


static int
sc_submitentry(char *entry)
{
	CURL *curl;
	int status;
	char *submission;
	int bufsize;
	sc_status_code ret;

	curl = curl_easy_init();
	curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1);
	curl_easy_setopt(curl, CURLOPT_URL, sc_submit_url);
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, sc_store_res);
	curl_easy_setopt(curl, CURLOPT_USERAGENT, USER_AGENT);
	curl_easy_setopt(curl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_0);

	pdebug(ht_fmt_vastr("Username: %s", sc_username), DEBUG);

#define SC_SUBMISSION_FORMAT "u=%s&s=%s%s" // entry begins '&'


	bufsize = strlen(SC_SUBMISSION_FORMAT) + strlen(sc_username) 
		+ strlen(sc_response_hash) + strlen(entry) + 1;

	submission = (char *)malloc(bufsize);

	snprintf(submission, bufsize, SC_SUBMISSION_FORMAT, sc_username,
			 sc_response_hash, entry);

	submission[bufsize - 1] = '\0';

	curl_easy_setopt(curl, CURLOPT_POSTFIELDS, submission);
	memset(sc_curl_errbuf, 0, sizeof(sc_curl_errbuf));
	curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, sc_curl_errbuf);
	curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, SC_CURL_TIMEOUT);

	if (sc_use_proxy) {
		curl_easy_setopt(curl, CURLOPT_PROXY, sc_proxy_server);

		if (sc_use_proxy_auth)
			curl_easy_setopt(curl, CURLOPT_PROXYUSERPWD, sc_proxy_userpwd);
	}

	sc_set_status(SB_CURL_CONNECT);
	sc_update_status(SB_CURL_CONNECT);

	pdebug("Now connecting...", DEBUG);

	status = DEBUG_SB_CURL ? CURLE_COULDNT_CONNECT
		: curl_easy_perform(curl);

	curl_easy_cleanup(curl);

	free(submission);

	if (status) {

		pdebug(sc_curl_errbuf, DEBUG);
		sc_sb_errors ++;
		sc_free_res();

		sc_set_status_curl(status);

		return SB_CURL_ERROR;
	}

	if ((ret = sc_parse_sb_res()) != SB_OK) {
		sc_sb_errors ++;
		sc_free_res();

		/*
		if (! sc_giveup)
			pdebug(ht_fmt_vastr("Retrying in %d secs, %d elements in queue",
								sc_submit_interval, hatena_queue_len()), 
				   DEBUG);
		*/

		return ret;
	}

	sc_free_res();

	return ret;
}


static void
write_cache(void)
{
	FILE *fd;
	sc_entry_t *entry;

	if (! sc_cachefile)
		return;

	if (! (fd = fopen(sc_cachefile, "w"))) {
		pdebug(ht_fmt_vastr("Failure opening %s", sc_cachefile), DEBUG);
		return;
	}

	pdebug(ht_fmt_vastr("Opening %s", sc_cachefile), DEBUG);

	hatena_queue_peek_all_rewind();

	while ((entry = hatena_queue_peek_all())) {
		pdebug(ht_fmt_vastr("a=%s t=%s l=%s i=%s m=%s b=%s",
							entry->artist ? entry->artist : "",
							entry->title ? entry->title : "",
							entry->len,
							entry->utctime ? entry->utctime : "",
							entry->mb ? entry->mb : "",
							entry->album ? entry->album : ""), DEBUG);

		fprintf(fd, "%s %s %s %s %s %s\n",
				entry->artist ? entry->artist : "",
				entry->title ? entry->title : "",
				entry->len,
				entry->utctime ? entry->utctime : "",
				entry->album ? entry->album : "",
				entry->mb ? entry->mb : "");
	}

	fclose(fd);
}


static void
sc_handlequeue(void)
{
	char *submitentry = NULL;
	int nsubmit = 0;
	int wait;
	sc_status_code ret;

	if (sc_sb_wait_until > time(NULL)) return ;

	//pdebug("mutex lock", DEBUG);
	pthread_mutex_lock(&sc_queue_mutex);
	
	if (hatena_queue_len() > 0)
		nsubmit = sc_generate_entry(&submitentry);

	pthread_mutex_unlock(&sc_queue_mutex);


	if (nsubmit > 0) {

		pdebug(ht_fmt_vastr("Number submitting: %d", nsubmit), DEBUG);
		pdebug(ht_fmt_vastr("Submission: %s", submitentry), DEBUG);

		if ((ret = sc_submitentry(submitentry)) != SB_OK) {

			if (sc_sb_errors < 5)
				wait = 60; /* Retry after 1 min */
			else if (sc_sb_errors < 12)
				wait = 60 << (sc_sb_errors - 5);
			else 
				wait = 60 << 7;
				
			sc_sb_wait_until = time(NULL) + wait;

			pdebug(ht_fmt_vastr("Error while submitting. "
								"Retrying after %d seconds.", wait),
				   DEBUG);
				
		} else {
			//pdebug("mutex lock", DEBUG);
			pthread_mutex_lock(&sc_queue_mutex);
			
			for ( ; nsubmit > 0; nsubmit --)
				sc_delete_entry(hatena_queue_pull());

			/*
			 * This should make sure that the queue doesn't
			 * get submitted multiple times on a nasty
			 * segfault...
			 */
			write_cache();

			pthread_mutex_unlock(&sc_queue_mutex);
			
			sc_sb_errors = 0;

			sc_sb_wait_until = time(NULL) + DEFAULT_SUBMIT_INTERVAL;
		}

		sc_set_status(ret);
			
	}

	if (submitentry)
		free(submitentry);

}


static void
read_cache(void)
{
	FILE *fd;
	int i = 0;
	char *linebuf = NULL;
	int linesize = 0;

#define LINE_SIZE 1024

	if (! sc_cachefile)
		return;

	if (! (fd = fopen(sc_cachefile, "r")))
		return;

	pdebug(ht_fmt_vastr("Opening %s", sc_cachefile), DEBUG);

	linesize = LINE_SIZE;
	linebuf = (char *)malloc(linesize);
	linebuf[0] = '\0';

	while (! feof(fd)) {
		char *artist, *title, *len, *utctime, *album, *mb;
		char *field[6] = {NULL};
		int j;
		sc_entry_t *entry;		

		fgets(linebuf + strlen(linebuf), linesize - strlen(linebuf), fd);
		if (linebuf[strlen(linebuf) - 1] != '\n') {
			linebuf = (char *)realloc(linebuf, linesize + LINE_SIZE);
			linesize += LINE_SIZE;
			continue;
		}

		linebuf[strlen(linebuf) - 1] = '\0';

		field[0] = linebuf;

		for (j = 1; j < 6; j ++) {
			char *p;
			if ((p = strchr(field[j - 1], ' '))) {
				*p = '\0';
				field[j] = p + 1;
			} else {
				pdebug("Cache format error:", DEBUG);
				break;
			}
		}

		artist = (strlen(field[0]) > 0) ? curl_unescape(field[0], 0) : NULL;
		title = (strlen(field[1]) > 0) ? curl_unescape(field[1], 0) : NULL;
		len = (strlen(field[2]) > 0) ? curl_unescape(field[2], 0) : NULL;
		utctime = (strlen(field[3]) > 0) ? curl_unescape(field[3], 0) : NULL;
		album = (strlen(field[4]) > 0) ? curl_unescape(field[4], 0) : NULL;
		mb = (strlen(field[5]) > 0) ? curl_unescape(field[5], 0) : NULL;

		hatena_queue_push((entry 
							  = sc_new_entry(artist, title, len, utctime, album, mb)));

		if (entry) {
			pdebug(ht_fmt_vastr("a[%d]=%s t[%d]=%s l[%d]=%s "
							 "i[%d]=%s m[%d]=%s b[%d]=%s",
							 i, entry->artist ? entry->artist : "",
							 i, entry->title ? entry->title : "",
							 i, entry->len,
							 i, entry->utctime ? entry->utctime : "",
							 i, entry->mb ? entry->mb : "",
							 i, entry->album ? entry->album : ""),
				   DEBUG);
		}
		
		if (artist) curl_free(artist);
		if (title) curl_free(title);
		if (utctime) curl_free(utctime);
		if (len) curl_free(len);
		if (album) curl_free(album);
		if (mb) curl_free(mb);

		i ++;
		
		linebuf[0] = '\0';
		
	}

	fclose(fd);


	pdebug("Done loading cache.", DEBUG);

	free(linebuf);
}



static void
sc_checkhandshake(void)
{
	int wait;
	sc_status_code ret;

	if (sc_hs_wait_until < time(NULL)) {

		if ((ret = sc_handshake()) != HS_OK) {

			if(sc_hs_errors < 5)
				/* Retry after 60 seconds */
				wait = 60;
			else if (sc_hs_errors < 12)
				wait =  60 << (sc_hs_errors - 5);
			else
				wait = 60 << 7;
			
			sc_hs_wait_until = time(NULL) + wait;

			pdebug(ht_fmt_vastr("Error while handshaking (%d). Retrying "
								"after %d seconds.", ret, wait), DEBUG);
		} else {
			sc_hs_status = 1;
			pdebug("handshake done", DEBUG);
		}

		sc_set_status(ret);
	}

}



/**** Public *****/

void
hatena_cleaner(void)
{
	sc_entry_t *entry;

	pdebug("cleaner", DEBUG);

	write_cache();

	hatena_queue_peek_all_rewind();

	while ((entry = (sc_entry_t *)hatena_queue_peek_all()))
		sc_delete_entry(hatena_queue_pull());

	hatena_reset_params();

	if (sc_cachefile) free (sc_cachefile);
	sc_cachefile = NULL;

	pdebug("hatena shutting down", DEBUG);
}


void
hatena_set_status_handler(void (* status_handler) (sc_status_code))
{
	cb_status_update = status_handler;
}

void
hatena_set_username(char *username)
{
	if (sc_username) 
		free(sc_username);
	sc_username = username ? strdup(username) : NULL;
}

void
hatena_set_password(char *password)
{
	if (sc_password)
		free(sc_password);
	sc_password = password ? strdup(password) : NULL;
}

void
hatena_set_proxy(int useproxy, char *server, char *port, 
					int useproxyauth, char *username, char *password)
{
	size_t bufsize;

	if (sc_proxy_server)
		free(sc_proxy_server);
	sc_proxy_server = NULL;

	if (sc_proxy_userpwd)
		free(sc_proxy_userpwd);
	sc_proxy_userpwd = NULL;

	sc_use_proxy = 0;
	sc_use_proxy_auth = 0;

	if (useproxy && server) {

		sc_use_proxy = 1;

		if (port) {
			bufsize = strlen(server) + sizeof(':') + strlen(port) + 1;
			sc_proxy_server = (char *)malloc(bufsize);
			snprintf(sc_proxy_server, bufsize, "%s:%s", server, port);
		} else {
			sc_proxy_server = strdup(server);
		}

		if (useproxyauth && username) {

			sc_use_proxy_auth = 1;

			bufsize = strlen(username) + sizeof(':') 
				+ strlen(password ? password : "") + 1;
			sc_proxy_userpwd = (char *)malloc(bufsize);
			/* Shouldn't we accept empty password? */
			snprintf(sc_proxy_userpwd, bufsize, "%s:%s", username, 
					 password ? password : "");
		}
	}
}

void
hatena_set_cachefile(const char *cache)
{
	if (sc_cachefile) free (sc_cachefile);
	sc_cachefile = cache ? strdup(cache) : NULL;
}

int
hatena_get_next_hs(void)
{
	return sc_hs_wait_until;
}

int
hatena_get_next_sb(void)
{
	return sc_sb_wait_until;
}

void
hatena_reset_params(void)
{
	sc_hs_status = 0;
	sc_hs_wait_until = 0;
	sc_hs_errors = 0;
	sc_sb_wait_until = 0;
	sc_srv_res_size = 0;
	sc_giveup = 0;
	sc_sb_errors = 0;
	/*sc_submit_interval = DEFAULT_SUBMIT_INTERVAL;*/

	sc_badauth_count = 0;
	
	sc_use_proxy = 0;
	sc_use_proxy_auth = 0;

	if (sc_submit_url) free(sc_submit_url);
	sc_submit_url = NULL;

	if (sc_username) free(sc_username);
	sc_username = NULL;
	
	if (sc_password) free(sc_password);
	sc_password = NULL;

	if (sc_srv_res) free(sc_srv_res);
	sc_srv_res = NULL;

	if (sc_challenge_hash) free(sc_challenge_hash);
	sc_challenge_hash = NULL;
	
	if (sc_proxy_server) free(sc_proxy_server);
	sc_proxy_server = NULL;

	if (sc_proxy_userpwd) free(sc_proxy_userpwd);
	sc_proxy_userpwd = NULL;

}


sc_status_code
hatena_get_status(void)
{
	sc_status_code ret;
	
	//pdebug("mutex lock", DEBUG);
	pthread_mutex_lock(&sc_status_mutex);

	ret = sc_status;

	pthread_mutex_unlock(&sc_status_mutex);

	return ret;
}

CURLcode
hatena_get_curl_status(void)
{
	CURLcode ret;

	//pdebug("mutex lock", DEBUG);
	pthread_mutex_lock(&sc_status_mutex);

	ret = sc_curl_status;

	pthread_mutex_unlock(&sc_status_mutex);

	return ret;

}


void
hatena_init_cache(char *cache_file)
{
	hatena_set_cachefile(cache_file);
	
	read_cache();
}



void
hatena_init(void)
{
	hatena_reset_params();

	pthread_mutex_init(&sc_status_mutex, NULL);
	pthread_mutex_init(&sc_queue_mutex, NULL);

	pdebug("hatena starting up", DEBUG);
}

int
hatena_add_entry(char *artist, char *title, time_t utctime,
					char *album, char *mb, int length)
{
	char len[16];
	sc_entry_t *entry;
	char utctime_str[20];

	if (! artist || ! title)
		return -1;

	strftime(utctime_str, sizeof(utctime_str), "%F %T", gmtime(&utctime));

	//pdebug("mutex lock", DEBUG);
	pthread_mutex_lock(&sc_queue_mutex);

	snprintf(len, sizeof(len), "%d", length);

	pdebug(ht_fmt_vastr("strlen(len) = %d", strlen(len)), DEBUG);

	entry = sc_new_entry(artist, title, len, utctime_str, album, mb);

	hatena_queue_push((void *)entry);

	/*
	 * This will help make sure the queue will be saved on a nasty
	 * segfault...
	 */
	write_cache();
	pthread_mutex_unlock(&sc_queue_mutex);

	return 1;
}


/* Call periodically from the plugin */
int
hatena_idle(void)
{

	if (! sc_hs_status)
		sc_checkhandshake();
	
	if (sc_hs_status)
		sc_handlequeue();
	
	if (sc_giveup)
		return -1;
	else
		return 0;
}
